from abc import abstractmethod
from typing import List, TYPE_CHECKING

from efootprint.abstract_modeling_classes.empty_explainable_object import EmptyExplainableObject
from efootprint.abstract_modeling_classes.explainable_object_dict import ExplainableObjectDict
from efootprint.abstract_modeling_classes.explainable_quantity import ExplainableQuantity
from efootprint.abstract_modeling_classes.source_objects import SourceValue
from efootprint.constants.units import u
from efootprint.core.hardware.hardware_base import HardwareBase

if TYPE_CHECKING:
    from efootprint.core.usage.edge_usage_pattern import EdgeUsagePattern


class EdgeHardware(HardwareBase):
    def __init__(self, name: str, carbon_footprint_fabrication: ExplainableQuantity, power: ExplainableQuantity,
                 lifespan: ExplainableQuantity):
        super().__init__(
            name, carbon_footprint_fabrication, power, lifespan, SourceValue(1 * u.dimensionless))
        self.unitary_power_per_usage_pattern = ExplainableObjectDict()
        self.nb_of_instances_per_usage_pattern = ExplainableObjectDict()
        self.instances_energy_per_usage_pattern = ExplainableObjectDict()
        self.energy_footprint_per_usage_pattern = ExplainableObjectDict()
        self.instances_fabrication_footprint_per_usage_pattern = ExplainableObjectDict()
        self.nb_of_instances = EmptyExplainableObject()
        self.instances_fabrication_footprint = EmptyExplainableObject()
        self.instances_energy = EmptyExplainableObject()
        self.energy_footprint = EmptyExplainableObject()

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self) -> List:
        return []

    @property
    def calculated_attributes(self):
        return ["nb_of_instances_per_usage_pattern", "instances_fabrication_footprint_per_usage_pattern",
                "unitary_power_per_usage_pattern", "instances_energy_per_usage_pattern",
                "energy_footprint_per_usage_pattern", "nb_of_instances", "instances_fabrication_footprint",
                "instances_energy", "energy_footprint"]

    @property
    def systems(self) -> List:
        return list(set(sum([elt.systems for elt in self.modeling_obj_containers], start=[])))

    @property
    @abstractmethod
    def edge_usage_patterns(self) -> List["EdgeUsagePattern"]:
        pass

    def update_dict_element_in_nb_of_instances_per_usage_pattern(self, usage_pattern: "EdgeUsagePattern"):
        self.nb_of_instances_per_usage_pattern[
            usage_pattern] = usage_pattern.nb_edge_usage_journeys_in_parallel.copy().set_label(
            f"Number of {self.name} instances for {usage_pattern.name}")

    def update_nb_of_instances_per_usage_pattern(self):
        self.nb_of_instances_per_usage_pattern = ExplainableObjectDict()
        for usage_pattern in self.edge_usage_patterns:
            self.update_dict_element_in_nb_of_instances_per_usage_pattern(usage_pattern)

    def update_dict_element_in_instances_fabrication_footprint_per_usage_pattern(
            self, usage_pattern: "EdgeUsagePattern"):
        instances_fabrication_footprint = (
                self.nb_of_instances_per_usage_pattern[usage_pattern] * self.carbon_footprint_fabrication
                * (ExplainableQuantity(1 * u.hour, "one hour") / self.lifespan))

        self.instances_fabrication_footprint_per_usage_pattern[usage_pattern] = instances_fabrication_footprint.to(
            u.kg).set_label(f"Hourly {self.name} instances fabrication footprint for {usage_pattern.name}")

    def update_instances_fabrication_footprint_per_usage_pattern(self):
        self.instances_fabrication_footprint_per_usage_pattern = ExplainableObjectDict()
        for usage_pattern in self.edge_usage_patterns:
            self.update_dict_element_in_instances_fabrication_footprint_per_usage_pattern(usage_pattern)

    @abstractmethod
    def update_unitary_power_per_usage_pattern(self):
        pass

    def update_dict_element_in_instances_energy_per_usage_pattern(self, usage_pattern: "EdgeUsagePattern"):
        unitary_energy = self.unitary_power_per_usage_pattern[usage_pattern] * ExplainableQuantity(1 * u.hour,
                                                                                                   "one hour")
        instances_energy = self.nb_of_instances_per_usage_pattern[usage_pattern] * unitary_energy

        self.instances_energy_per_usage_pattern[usage_pattern] = instances_energy.set_label(
            f"Hourly energy consumed by {self.name} instances for {usage_pattern.name}")

    def update_instances_energy_per_usage_pattern(self):
        self.instances_energy_per_usage_pattern = ExplainableObjectDict()
        for usage_pattern in self.edge_usage_patterns:
            self.update_dict_element_in_instances_energy_per_usage_pattern(usage_pattern)

    def update_dict_element_in_energy_footprint_per_usage_pattern(self, usage_pattern: "EdgeUsagePattern"):
        energy_footprint = (self.instances_energy_per_usage_pattern[usage_pattern] *
                            usage_pattern.country.average_carbon_intensity)
        self.energy_footprint_per_usage_pattern[usage_pattern] = energy_footprint.set_label(
            f"{self.name} energy footprint for {usage_pattern.name}")

    def update_energy_footprint_per_usage_pattern(self):
        self.energy_footprint_per_usage_pattern = ExplainableObjectDict()
        for usage_pattern in self.edge_usage_patterns:
            self.update_dict_element_in_energy_footprint_per_usage_pattern(usage_pattern)

    def sum_calculated_attribute_across_usage_patterns(self, calculated_attribute_name: str,
                                                       calculated_attribute_label: str):
        summed_attribute = EmptyExplainableObject()
        for usage_pattern in self.edge_usage_patterns:
            summed_attribute += getattr(self, calculated_attribute_name)[usage_pattern]

        return summed_attribute.set_label(f"{self.name} {calculated_attribute_label} across usage patterns")

    def update_nb_of_instances(self):
        self.nb_of_instances = self.sum_calculated_attribute_across_usage_patterns(
            "nb_of_instances_per_usage_pattern", "total instances")

    def update_instances_energy(self):
        self.instances_energy = self.sum_calculated_attribute_across_usage_patterns(
            "instances_energy_per_usage_pattern", "total instances energy")

    def update_energy_footprint(self):
        self.energy_footprint = self.sum_calculated_attribute_across_usage_patterns(
            "energy_footprint_per_usage_pattern", "total energy footprint")

    def update_instances_fabrication_footprint(self):
        self.instances_fabrication_footprint = self.sum_calculated_attribute_across_usage_patterns(
            "instances_fabrication_footprint_per_usage_pattern", "total fabrication footprint")



from typing import List, TYPE_CHECKING, Optional
import numpy as np
from pint import Quantity

from efootprint.abstract_modeling_classes.explainable_object_dict import ExplainableObjectDict
from efootprint.abstract_modeling_classes.explainable_recurrent_quantities import ExplainableRecurrentQuantities
from efootprint.abstract_modeling_classes.modeling_object import ModelingObject
from efootprint.abstract_modeling_classes.source_objects import SourceRecurrentValues
from efootprint.constants.units import u

if TYPE_CHECKING:
    from efootprint.core.usage.edge_usage_pattern import EdgeUsagePattern
    from efootprint.core.usage.edge_usage_journey import EdgeUsageJourney
    from efootprint.core.hardware.edge_device import EdgeDevice
    from efootprint.core.system import System


class RecurrentEdgeProcess(ModelingObject):
    default_values = {
        "recurrent_compute_needed": SourceRecurrentValues(Quantity(np.array([1] * 168, dtype=np.float32), u.cpu_core)),
        "recurrent_ram_needed": SourceRecurrentValues(Quantity(np.array([1] * 168, dtype=np.float32), u.GB)),
        "recurrent_storage_needed": SourceRecurrentValues(Quantity(np.array([0] * 168, dtype=np.float32), u.GB)),
    }

    def __init__(self, name: str, recurrent_compute_needed: ExplainableRecurrentQuantities, 
                 recurrent_ram_needed: ExplainableRecurrentQuantities,
                 recurrent_storage_needed: ExplainableRecurrentQuantities):
        super().__init__(name)
        self.unitary_hourly_compute_need_per_usage_pattern = ExplainableObjectDict()
        self.unitary_hourly_ram_need_per_usage_pattern = ExplainableObjectDict()
        self.unitary_hourly_storage_need_per_usage_pattern = ExplainableObjectDict()
        
        self.recurrent_compute_needed = recurrent_compute_needed.set_label(f"{self.name} recurrent compute needed")
        self.recurrent_ram_needed = recurrent_ram_needed.set_label(f"{self.name} recurrent ram needed")
        self.recurrent_storage_needed = recurrent_storage_needed.set_label(f"{self.name} recurrent storage needed")

    @property
    def calculated_attributes(self):
        return ["unitary_hourly_compute_need_per_usage_pattern", "unitary_hourly_ram_need_per_usage_pattern",
                "unitary_hourly_storage_need_per_usage_pattern"]

    @property
    def edge_usage_journeys(self) -> List["EdgeUsageJourney"]:
        return self.modeling_obj_containers

    @property
    def edge_usage_patterns(self) -> List["EdgeUsagePattern"]:
        return list(set(sum([euj.edge_usage_patterns for euj in self.edge_usage_journeys], start=[])))

    @property
    def edge_devices(self) -> List["EdgeDevice"]:
        return [euj.edge_device for euj in self.edge_usage_journeys]
        
    @property
    def systems(self) -> List["System"]:
        return list(set(sum([euj.systems for euj in self.edge_usage_journeys], start=[])))

    @property
    def modeling_objects_whose_attributes_depend_directly_on_me(self) -> List:
        return []

    def update_dict_element_in_unitary_hourly_compute_need_per_usage_pattern(self, usage_pattern: "EdgeUsagePattern"):
        unitary_hourly_compute_need = self.recurrent_compute_needed.generate_hourly_quantities_over_timespan(
            usage_pattern.hourly_edge_usage_journey_starts, usage_pattern.country.timezone)
        self.unitary_hourly_compute_need_per_usage_pattern[usage_pattern] = unitary_hourly_compute_need.set_label(
            f"{self.name} unitary hourly compute need for {usage_pattern.name}")

    def update_unitary_hourly_compute_need_per_usage_pattern(self):
        self.unitary_hourly_compute_need_per_usage_pattern = ExplainableObjectDict()
        for usage_pattern in self.edge_usage_patterns:
            self.update_dict_element_in_unitary_hourly_compute_need_per_usage_pattern(usage_pattern)

    def update_dict_element_in_unitary_hourly_ram_need_per_usage_pattern(self, usage_pattern: "EdgeUsagePattern"):
        unitary_hourly_ram_need = self.recurrent_ram_needed.generate_hourly_quantities_over_timespan(
            usage_pattern.hourly_edge_usage_journey_starts, usage_pattern.country.timezone)
        self.unitary_hourly_ram_need_per_usage_pattern[usage_pattern] = unitary_hourly_ram_need.set_label(
            f"{self.name} unitary hourly ram need for {usage_pattern.name}")

    def update_unitary_hourly_ram_need_per_usage_pattern(self):
        self.unitary_hourly_ram_need_per_usage_pattern = ExplainableObjectDict()
        for usage_pattern in self.edge_usage_patterns:
            self.update_dict_element_in_unitary_hourly_ram_need_per_usage_pattern(usage_pattern)

    def update_dict_element_in_unitary_hourly_storage_need_per_usage_pattern(self, usage_pattern: "EdgeUsagePattern"):
        unitary_hourly_storage_need = self.recurrent_storage_needed.generate_hourly_quantities_over_timespan(
            usage_pattern.hourly_edge_usage_journey_starts, usage_pattern.country.timezone)
        self.unitary_hourly_storage_need_per_usage_pattern[usage_pattern] = unitary_hourly_storage_need.set_label(
            f"{self.name} unitary hourly storage need for {usage_pattern.name}")

    def update_unitary_hourly_storage_need_per_usage_pattern(self):
        self.unitary_hourly_storage_need_per_usage_pattern = ExplainableObjectDict()
        for usage_pattern in self.edge_usage_patterns:
            self.update_dict_element_in_unitary_hourly_storage_need_per_usage_pattern(usage_pattern)
